use std::path::{Path, PathBuf};

use futures::stream::TryStreamExt;
use reqwest::{
    header::{HeaderMap, HeaderName, HeaderValue},
    Method as ReqMethod,
};
use rocket::{
    http::uri::Segments,
    fs::Options,
    data::ToByteUnit,
    http::{Method, Status},
    route::{Handler, Outcome},
    Data, Request, Route,
};
use tokio_util::compat::FuturesAsyncReadCompatExt;

#[derive(Clone)]
pub struct ProxyServer
{
    root: String,
    prefix: PathBuf,
    rank: isize,
    options: Options,
}

impl ProxyServer
{
    const DEFAULT_RANK: isize = 0;

    pub fn new<P: AsRef<Path>>(root: &str, prefix: P, rank: isize, options: Options) -> Self
    {
        ProxyServer {
            root: String::from(root),
            prefix: prefix.as_ref().into(),
            rank,
            options,
        }
    }

    pub fn from<P: AsRef<Path>>(root: &str, prefix: P) -> Self
    {
        Self::new(root, prefix, Self::DEFAULT_RANK, Options::NormalizeDirs)
    }
}

impl Into<Vec<Route>> for ProxyServer
{
    fn into(self) -> Vec<Route>
    {
        let mut routes: Vec<Route> = Vec::new();

        static METHODS: [Method; 5] = [
            Method::Get,
            Method::Post,
            Method::Patch,
            Method::Delete,
            Method::Put,
        ];

        for method in METHODS {
            let mut route = Route::ranked(self.rank, method, "/<path..>", self.clone());
            route.name = Some(
                format!(
                    "ProxyServer: {} {}{}",
                    method.as_str(),
                    self.root.clone(),
                    self.prefix.to_str().unwrap()
                )
                .into(),
            );

            routes.push(route);
        }

        routes
    }
}

#[rocket::async_trait]
impl Handler for ProxyServer
{
    async fn handle<'r>(&self, req: &'r Request<'_>, data: Data<'r>) -> Outcome<'r>
    {
        use rocket::http::uri::fmt::Path;

        // ROCKET REQUEST -> REQWEST REQUEST
        // Path
        let allow_dotfiles = self.options.contains(Options::DotFiles);
        let path = req.segments::<Segments<'_, Path>>(0..).ok()
            .and_then(|segments| segments.to_path_buf(allow_dotfiles).ok())
            .map(|path| self.prefix.join(path));

        if path.is_none() {
            return Outcome::Failure(Status::BadRequest);
        }

        let path_val = path.unwrap();
        let path_str = path_val.to_str(); // Unwrap is safe because we check for is_none() beforehand

        if path_str.is_none() {
            return Outcome::Failure(Status::BadRequest);
        }

        let query_part = req
            .uri()
            .query()
            .map_or(String::from(""), |s| format!("?{}", s));

        let url = format!(
            "{}{}{}",
            self.root,
            path_str.unwrap(),  // Unwrap is safe because we check for is_none() beforehand
            query_part
        );

        // Headers
        let mut headers = HeaderMap::new();

        for header in req.headers().iter() {
            let name = match HeaderName::from_lowercase(header.name.into_string().as_bytes()) {
                Ok(val) => val,
                Err(_) => return Outcome::Failure(Status::BadRequest),
            };
            let value = match HeaderValue::from_str(&header.value.into_owned()) {
                Ok(val) => val,
                Err(_) => return Outcome::Failure(Status::BadRequest),
            };

            headers.insert(name, value);
        }

        // Method
        let method = match req.method() {
            Method::Delete => ReqMethod::DELETE,
            Method::Get => ReqMethod::GET,
            Method::Patch => ReqMethod::PATCH,
            Method::Post => ReqMethod::POST,
            Method::Put => ReqMethod::PUT,
            _ => todo!(),
        };

        // And finally, the data
        // TODO don't hard-code max request size here
        static BODY_SIZE: i64 = 2;
        // TODO remove unwrap
        let body: Vec<u8> = data
            .open(BODY_SIZE.mebibytes())
            .into_bytes()
            .await
            .unwrap()
            .to_vec();

        // TODO remove unwrap
        // let body = reqwest::Body::wrap_stream(data);
        let new_res = reqwest::Client::new()
            .request(method, url)
            .headers(headers)
            .body(body)
            .send()
            .await
            .unwrap();
        let mut res_headers = new_res.headers().clone();

        let mut res_builder = rocket::Response::build();

        // Let us thank this Git repo for showing me how to do this
        // https://github.com/benkay86/async-applied/blob/master/reqwest-tokio-compat/src/main.rs
        let mut res_builder = res_builder
            .status(rocket::http::Status::new(new_res.status().as_u16()))
            .streamed_body(
                new_res
                    .bytes_stream()
                    .map_err(|e| futures::io::Error::new(futures::io::ErrorKind::Other, e))
                    .into_async_read()
                    .compat(),
            );

        // reqwest headers -> rocket headers
        for (key, value) in res_headers.iter_mut() {
            res_builder = res_builder.raw_header(
                String::from(key.clone().as_str()),
                String::from(value.clone().to_str().unwrap()),
            );
        }

        Outcome::Success(res_builder.finalize())
    }
}
